from enum import Enum

class Answer(Enum):
    YES = 1
    NO = 2

def question(q: str, a: Answer) -> str:
    """
    Answer to a question

    q: The question
    a: The answer
    """
    pass

def add(x: int, y: int) -> int:
    """
    An adder function that allows for all types.    

    x: The first number to add.
    y: The second number to add.
    """
    return x + y

def multiply(x: int, y: int) -> int:
    """
    A function for muliplying two integers.

    x: The first integer
    y: The second integer
    """
    return x * y

def concat(first: str, second: str) -> str:
    """
    A leetle function to concatentate strings

    first: The first string
    second: The second string

    returns concatenated string
    """
    return first + second




import base64
from typing import get_type_hints, Callable, Dict, Any
from enum import Enum
from pathlib import Path
import json
import re
from toolla.exceptions import InvalidDescriptionException

def extract_json_from_text(text):
    text_without_newlines = text.replace('\n', '')
    start_index = None
    curly_brace_count = 0
    for i, c in enumerate(text_without_newlines):
        if c == '{':
            curly_brace_count += 1
            if not start_index:
                start_index = i
        if c == '}':
            curly_brace_count -= 1
            if curly_brace_count == 0:
                return json.loads(text_without_newlines[start_index:i+1])
    return None

# Generated by Claude Sonnet 3.5
def parse_descriptions(input_string):
    lines = input_string.strip().split('\n')
    result = {}
    colon_found = False
    for i, line in enumerate(lines):
        if i == 0:
            result['fn_description'] = line.strip()
        elif ':' in line:
            colon_found = True
            key, value = line.split(':', 1)
            key = key.strip()
            value = value.strip()
            result[key] = value
    
    if not colon_found:
        raise InvalidDescriptionException
    
    return result

def build_openai_tool_schema(f: Callable):
    """Build a JSON schema for an OpenAI tool function."""
    descriptions = parse_descriptions(f.__doc__)
    hints = get_type_hints(f)
    schema = {
        "type": "function",
        "function": {
            "name": f.__name__,
            "description": descriptions['fn_description'],
            "parameters": {
                "type": "object",
                "properties": {},
                "required": []
            },
        }
    }

    for param, hint in hints.items():
        if param == "return":
            continue
        if isinstance(hint, type) and issubclass(hint, Enum):
            schema["function"]["parameters"]["properties"][param] = {
                "type": "string",
                "enum": [e.name for e in hint],
                "description": descriptions[param]
            }
            schema["function"]["parameters"]["required"].append(param)
            continue

        var_type = "number" if hint is int or hint is float else "string"
        schema["function"]["parameters"]["properties"][param] = {
            "type": var_type,
            "description": descriptions[param]
        }
        schema["function"]["parameters"]["required"].append(param)
    return schema

# TODO add support for enums
def build_claude_tool_schema(f: Callable):
    """Build a JSON schema for a tool function."""
    descriptions = parse_descriptions(f.__doc__)
    hints = get_type_hints(f)
    schema = {
        "name": f.__name__,
        "description": descriptions['fn_description'],
        "input_schema": {
            "type": "object",
            "properties": {},
            "required": []
        },
    }

    for param, hint in hints.items():
        if param == "return":
            continue
        if isinstance(hint, type) and issubclass(hint, Enum):
            schema["input_schema"]["properties"][param] = {
                "type": "string",
                "enum": [e for e in hint.__members__],
                "description": descriptions[param]
            }
            schema["input_schema"]["required"].append(param)
            continue

        var_type = "number" if hint is int or hint is float else "string"
        schema["input_schema"]["properties"][param] = {
            "type": var_type,
            "description": descriptions[param]
        }
        schema["input_schema"]["required"].append(param)
    return schema

def load_file_base64(file_path: Path) -> str:
    with open(file_path, "rb") as file:
        file_content = file.read()
        return base64.b64encode(file_content).decode('utf-8')

def get_image_mime_type(fpath: Path) -> str:
    suffix = fpath.suffix
    mime_types = {
        '.jpg': 'image/jpeg',
        '.jpeg': 'image/jpeg',
        '.png': 'image/png',
        '.gif': 'image/gif',
        '.webp': 'image/webp',
    }
    return mime_types.get(suffix)



from typing import Union, List, Callable
from openai import OpenAI
from toolla.exceptions import (
    ImageNotSupportedException,
    AbortedToolException,
    MessageTooLongException,
)
from toolla.utils import (
    build_openai_tool_schema,
    extract_json_from_text,
)
from toolla.models import default_tool_prompt

class OpenAICompatibleClient:
    def __init__(
        self,
        model: str,
        *,
        tools: List[Callable] = [],
        max_steps = 10,
        print_output=False,
        base_url: Union[str, None] = None,
        system: Union[str, None] = None,
    ):
        self.client = OpenAI(
            base_url=base_url,
        )
        self.model = model
        self.max_steps = max_steps
        self.messages = []
        self.print_output = print_output
        self.max_chars = 900_000 # Totally random, figure something else out

        if tools:
            self.tools = []
            self.tool_fns = {}
            for f in tools:
                self.tool_fns[f.__name__] = f
                tool_dict = build_openai_tool_schema(f)
                self.tools.append(tool_dict)
        else:
            self.tools = []

        self.messages.append(
            {
                "role": "system",
                "content": system or default_tool_prompt.format(tool_list=str(self.tools)),
            }
        )
        
    def __call__(
        self,
        prompt: str,
        image: Union[str, None] = None,
        current_fn_response = None,
        disable_auto_execution = False,
    ):
        message = {
            "role": "user",
            "content": prompt,
        }
        self.messages.append(message)
        if image:
            raise ImageNotSupportedException
    
        while len(str(self.messages)) > self.max_chars:
            self.messages.pop(0)
            if not self.messages:
                raise MessageTooLongException

        response = self.client.chat.completions.create(
            model=self.model,
            messages=self.messages,
        )
        self.messages.append({
            "role": "assistant",
            "content": response.choices[0].message.content,
        })
        if self.print_output:
            print(f"{response.choices[0].message.content}\n")
        parsed_response = extract_json_from_text(response.choices[0].message.content)
        if parsed_response:
            if disable_auto_execution:
                print(f"Function {parsed_response['tool']} is about to be called with inputs: {parsed_response['inputs']}")
                user_input = input("Do you want to run this function? (y/n): ")
                if user_input.lower() not in ['y', 'Y']:
                    print("Function call aborted by user.")
                    raise AbortedToolException
            r = self.tool_fns[parsed_response['tool']](**parsed_response['inputs'])
            if len(self.messages) < 2 * self.max_steps:
                return self(
                    prompt=f"\nFunction {parsed_response['tool']} was called and returned a value of {r}",
                    current_fn_response=r,
                    disable_auto_execution=disable_auto_execution,
                )
            else:
                print("Reached maxiumum number of steps, returning current tool response.")
                return current_fn_response
        else:
            return current_fn_response


import pytest
import random
import string
import builtins
from toolla.chat import Chat
from toolla.exceptions import (
    MessageTooLongException,
    AbortedToolException,
    ImageNotSupportedException,
)
from .tools import add, multiply, concat

def test_openai_compatible_client_clear_messages():
    chat = Chat(
        model="llama3.1",
        base_url="http://localhost:11434/v1",
        tools=[add],
    )
    r = chat("Hello")
    chat.clear_messages()
    messages = chat.get_messages()
    assert len(messages) == 1
    assert messages[0]["role"] == "system"

def test_ollama_llama3_add_tool():
    chat = Chat(
        model="llama3.1",
        base_url="http://localhost:11434/v1",
        tools=[add],
    )
    r = chat("What is 341+18?")
    assert r == 359

def test_llama_concat_tool():
    chat = Chat(
        model="llama3.1",
        base_url="http://localhost:11434/v1",
        tools=[concat]
    )
    r = chat("Concatenate the strings dream and tremor")
    assert r == "dreamtremor"

def test_llama_image_not_supported():
    with pytest.raises(ImageNotSupportedException) as exc_info:
        chat = Chat(
            model="llama3.1",
            base_url="http://localhost:11434/v1",
        )
        chat(prompt="What is this an image of? Answer with one word.", image="./cat.jpg")
    assert str(exc_info.value) == "Error: Image content not supported by API."

def test_llama_multiple_tools():
    chat = Chat(
        model="llama3.1",
        base_url="http://localhost:11434/v1",
        tools=[add, multiply],
    )
    r = chat("What is (4*4911)+18?")
    assert r == 19662

def test_llama_large_message_fail():
    with pytest.raises(MessageTooLongException) as exc_info:
        characters = string.ascii_letters + string.digits + string.punctuation
        prompt = ''.join(random.choice(characters) for _ in range(2000000))
        chat = Chat(
            model="llama3.1",
            base_url="http://localhost:11434/v1",
        )
        r = chat(prompt=prompt)
    assert str(exc_info.value) == "Error: Message is too long"

def test_llama_disable_auto_execution_answer_no():
    with pytest.raises(AbortedToolException) as exc_info:
        builtins.input = lambda _: 'n'
        chat = Chat(
            model="llama3.1",
            base_url="http://localhost:11434/v1",
            tools=[add],
        )
        chat("What is 4911+18?", disable_auto_execution=True)
    assert str(exc_info.value) == "Error: User aborted tool use."



import os
import json
from typing import Union, List, Callable
from pathlib import Path
from openai import OpenAI

from toolla.utils import (
    get_image_mime_type, 
    load_file_base64, 
    build_openai_tool_schema,
)
from toolla.exceptions import (
    MessageTooLongException, 
    AbortedToolException
)

class OpenAIClient:
    def __init__(
        self,
        model: str = "gpt-4o",
        system: Union[str, None] = None,
        tools: List[Callable] = [],
        max_steps: int = 10,
        print_output: bool = False,
        api_key: Union[str, None] = None,
    ):
        self.client = OpenAI(api_key=api_key or os.environ.get("OPENAI_API_KEY"))
        self.model = model
        self.max_steps = max_steps
        self.messages = []
        self.print_output = print_output

        # TODO change based on model choice
        self.max_tokens = 4096
        self.max_chars = 900_000

        if system:
            self.messages.append(
                {
                    "role": "system",
                    "content": system,
                }
            )

        if tools:
            # TODO check that docstrings
            # match functions and are valid 
            self.tools = []
            self.tool_fns = {}
            for f in tools:
                self.tool_fns[f.__name__] = f
                tool_dict = build_openai_tool_schema(f)
                self.tools.append(tool_dict)
        else:
            self.tools = []
    
    def __call__(
        self,
        prompt: str,
        image: Union[str, None] = None,
        current_fn_response = None,
        disable_auto_execution = False,
    ):
        message = {
            "role": "user",
            "content": [
                { "type": "text", "text": prompt }
            ]
        }
        self.messages.append(message)
        if image:
            fpath = Path(image)
            mtype = get_image_mime_type(fpath)
            image_string = load_file_base64(fpath) 
            message["content"].append(
                {
                    "type": "image_url",
                    "image_url": {
                        "url": f"data:{mtype};base64,{image_string}"
                    }
                }
            )
    
        while len(str(self.messages)) > self.max_chars:
            self.messages.pop(0)
            if not self.messages:
                raise MessageTooLongException

        # OpenAI doesn't allow for empty tool list
        # TODO change to self.tools or None
        response = self.client.chat.completions.create(
            model=self.model,
            max_tokens=self.max_tokens,
            messages=self.messages,
            tools=self.tools,
        ) if self.tools else self.client.chat.completions.create(
            model=self.model,
            max_tokens=self.max_tokens,
            messages=self.messages,
        )
 
        for choice in response.choices:
            if choice.finish_reason == 'stop':
                # TODO should be moved outside of for loop?
                self.messages.append({
                    "role": "assistant",
                    "content": choice.message.content,
                })
                if self.print_output:
                    print(f"{choice.message.content}\n")
                return current_fn_response
            elif choice.finish_reason == 'tool_calls':
                function = choice.message.tool_calls[0].function
                if disable_auto_execution:
                    print(f"Function {function.name} is about to be called with inputs: {function.arguments}")
                    user_input = input("Do you want to run this function? (y/n): ")
                    if user_input.lower() not in ['y', 'Y']:
                        print("Function call aborted by user.")
                        raise AbortedToolException
                r = self.tool_fns[function.name](**json.loads(function.arguments))
                if len(self.messages) < 2 * self.max_steps:
                    return self(
                        prompt=f"\nFunction {function.name} was called and returned a value of {r}",
                        current_fn_response=r,
                        disable_auto_execution=disable_auto_execution,
                    )
                else:
                    print("Reached maxiumum number of steps, returning current tool response.")
                    return current_fn_response


import pytest
import base64
from pathlib import Path
from toolla.utils import (
    build_claude_tool_schema,
    build_openai_tool_schema,
    load_file_base64,
    get_image_mime_type,
    parse_descriptions,
    extract_json_from_text,
)
from toolla.exceptions import InvalidDescriptionException
from .tools import add, question, multiply

def test_extract_json_block_fail():
    s = 'nclude the calculation.\n\nThe final answer is 13544.'
    parsed_s = extract_json_from_text(s)
    assert parsed_s == None

def test_extract_json_succeed():
    s = ' and inputs to use: {\n  "tool": "multiply",\n  "inputs": {\n    "x": 1313,\n    "y": 10\n  }\n} Please use the "multiply" tool with the inputs above, and then add the result to the conversation. I\'ll then guide you on the next step.'
    parsed_s = extract_json_from_text(s)
    assert parsed_s['tool'] == 'multiply'
    assert parsed_s['inputs']['x'] == 1313
    assert parsed_s['inputs']['y'] == 10

def test_extract_json_block_succeed_with_json_tag():
    s = ' and inputs to use:\n\n```json\n{\n  "tool": "multiply",\n  "inputs": {\n    "x": 1313,\n    "y": 10\n  }\n}\n```\n\nPlease use the "multiply" tool with the inputs above, and then add the result to the conversation. I\'ll then guide you on the next step.'
    parsed_s = extract_json_from_text(s)
    assert parsed_s['tool'] == 'multiply'
    assert parsed_s['inputs']['x'] == 1313
    assert parsed_s['inputs']['y'] == 10

def test_extract_json_block_succeed_without_json_tag():
    s = ' and inputs to use:\n\n```\n{\n  "tool": "multiply",\n  "inputs": {\n    "x": 1313,\n    "y": 10\n  }\n}\n```\n\nPlease use the "multiply" tool with the inputs above, and then add the result to the conversation. I\'ll then guide you on the next step.'
    parsed_s = extract_json_from_text(s)
    assert parsed_s['tool'] == 'multiply'
    assert parsed_s['inputs']['x'] == 1313
    assert parsed_s['inputs']['y'] == 10

def test_parse_description_succeed():
    descriptions = parse_descriptions(add.__doc__)
    assert descriptions['fn_description'] == "An adder function that allows for all types."
    assert descriptions['x'] == "The first number to add."
    assert descriptions['y'] == "The second number to add."

def test_failed_description_parse():
    with pytest.raises(InvalidDescriptionException) as excinfo:
        description = """
        This is a test description
        """
        descriptions = parse_descriptions(description)
    assert str(excinfo.value) == "Error: Invalid description."

def test_build_openai_tool_schema():
    schema = build_openai_tool_schema(add)
    expected = {
        "type": "function",
        "function": {
            "name": "add",
            "description": "An adder function that allows for all types.",
            "parameters": {
                "type": "object",
                "properties": {
                    "x": {
                        "type": "number",
                        "description": "The first number to add."
                    },
                    "y": {
                        "type": "number",
                        "description": "The second number to add."
                    },
                },
                "required": ["x", "y"]
            },
        }
    }
    assert schema == expected

def test_build_openai_tool_schema_with_enum():
    schema = build_openai_tool_schema(question)
    expected = {
        "type": "function",
        "function": {
            "name": "question",
            "description": "Answer to a question",
            "parameters": {
                "type": "object",
                "properties": {
                    "q": {
                        "type": "string",
                        "description": "The question"
                    },
                    "a": {
                        "type": "string",
                        "enum": ["YES", "NO"],
                        "description": "The answer"
                    }
                },
                "required": ["q", "a"]
            },
        }
    }
    assert schema == expected

def test_build_claude_tool_schema():
    schema = build_claude_tool_schema(add)
    expected = {
        "name": "add",
        "description": "An adder function that allows for all types.",
        "input_schema": {
            "type": "object",
            "properties": {
                "x": {
                    "type": "number",
                    "description": "The first number to add."
                },
                "y": {
                    "type": "number",
                    "description": "The second number to add."
                },
            },
            "required": ["x", "y"]
        },
    }
    assert schema == expected

def test_build_claude_schema_with_enum():
    schema = build_claude_tool_schema(question)
    expected = {
        "name": "question",
        "description": "Answer to a question",
        "input_schema": {
            "type": "object",
            "properties": {
                "q": {
                    "type": "string",
                    "description": "The question"
                },
                "a": {
                    "type": "string",
                    "enum": ["YES", "NO"],
                    "description": "The answer"
                }
            },
            "required": ["q", "a"]
        },
    }
    assert schema == expected

def test_load_file_base64():
    # Create a temporary file with some content
    test_file = Path("test_file.txt")
    test_content = b"Hello, World!"
    test_file.write_bytes(test_content)

    try:
        # Test the function
        result = load_file_base64(test_file)
        expected = base64.b64encode(test_content).decode('utf-8')
        assert result == expected
    finally:
        # Clean up the temporary file
        test_file.unlink()

def test_get_jpeg_mime():
    fpath = Path('arrakis.jpg')
    t = get_image_mime_type(fpath)
    assert t == "image/jpeg"

def test_get_jpeg_alt_mime():
    fpath = Path('dune.jpeg')
    t = get_image_mime_type(fpath)
    assert t == "image/jpeg"

def test_get_gif_mime():
    fpath = Path('sandworm.gif')
    t = get_image_mime_type(fpath)
    assert t == "image/gif"

def test_get_png_mime():
    fpath = Path('fremen.png')
    t = get_image_mime_type(fpath)
    assert t == "image/png"

def test_get_webp_mime():
    fpath = Path('spice.webp')
    t = get_image_mime_type(fpath)
    assert t == "image/webp"



class MessageTooLongException(Exception):
    def __init__(self, message="Error: Message is too long"):
        self.message = message
        super().__init__(self.message)

class AbortedToolException(Exception):
    def __init__(self, message="Error: User aborted tool use."):
        self.message = message
        super().__init__(self.message)


class ModelNotSupportedException(Exception):
    def __init__(self, message="Error: Model not supported by library."):
        self.message = message
        super().__init__(self.message)

class ImageNotSupportedException(Exception):
    def __init__(self, message="Error: Image content not supported by API."):
        self.message = message
        super().__init__(self.message)

class InvalidDescriptionException(Exception):
    def __init__(self, message="Error: Invalid description."):
        self.message = message
        super().__init__(self.message)


import pytest
import os
import builtins
from toolla.chat import Chat
from toolla.exceptions import (
    MessageTooLongException,
    AbortedToolException,
)
from toolla.models import models
from .tools import add, multiply, concat

def test_openai_client_clear_message():
    chat = Chat(model="gpt-4o")
    chat("Hello")
    chat.clear_messages()
    assert len(chat.get_messages()) == 0

def test_openai_add_tool():
    chat = Chat(model="gpt-4o", tools=[add])
    r = chat("What is 2+3?")
    assert r == 5

def test_openai_concat_tool():
    chat = Chat(model="gpt-4o", tools=[concat])
    r = chat("Concatenate the strings dream and tremor")
    assert r == "dreamtremor"

def test_openai_image_content():
    chat = Chat(model="gpt-4o")
    chat(prompt="What is this an image of? Answer with one word.", image="./tests/cat.jpg")
    messages = chat.get_messages()
    assert 'Cat' in messages[-1]['content'] or 'cat' in messages[-1]['content']

def test_openai_multiple_tools():
    chat = Chat(model="gpt-4o", tools=[add, multiply])
    r = chat("What is (4*4911)+18?")
    assert r == 19662

def test_openai_large_message_fail():
    try:
        with open('image.jpeg', 'wb') as f:
            f.write(b'\0' * 2000000)
        chat = Chat(model='gpt-4o')
        r = chat(prompt="What is this?", image='./image.jpeg')
    except MessageTooLongException as e:
        assert str(e) == "Error: Message is too long"
    finally:
        if os.path.exists('image.jpeg'):
            os.remove('image.jpeg')

def test_openai_disable_auto_execution_answer_no():
    with pytest.raises(AbortedToolException) as exc_info:
        builtins.input = lambda _: 'n'
        chat = Chat(model="gpt-4o", tools=[add])
        chat("What is (4*4911)+18?", disable_auto_execution=True)
    assert str(exc_info.value) == "Error: User aborted tool use."


default_tool_prompt = """
You are a helpful assistant that tells the user which tools to use and which inputs to give to the tools.  The list of descriptions of available tools in JSON format is 
{tool_list}.  
The user will ask a task to be performed and you will respond with the proper tool to use.  If a task requires multiple steps you will return the tool for the first step and the user will respond with an result from the tool used. If no tool is needed, simply answer the question without it. You will return tool to be used in  the JSON format with following structure 
```json
{{ 
    "tool": "<name>", 
    "inputs": {{ 
        "<key>": "<value>"
    }} 
}}
```
"""

# TODO add all subsequent models
models = {
    "openai_models": [
        "gpt-4o",
        "gpt-4o-2024-05-13",
        "gpt-4o-mini",
        "gpt-4o-mini-2024-07-18",
        "gpt-4-turbo",
        "gpt-4-turbo-2024-04-09",
    ],
    "claude_models": [
        "claude-3-5-sonnet-20240620",
        "claude-3-opus-20240229",
        "claude-3-sonnet-20240229",
        "claude-3-haiku-20240307",
    ],
}



from typing import List, Dict, Union, Callable
from toolla.models import (
    models,
)
from toolla.exceptions import (
    ModelNotSupportedException,
)
from toolla.anthropic_client import AnthropicClient
from toolla.openai_client import OpenAIClient
from toolla.openai_compatible_client import OpenAICompatibleClient

class Chat:
    def __init__(
        self, 
        model: str = "claude-3-5-sonnet-20240620",
        system: Union[str, None] = None,
        tools: List[Callable] = [],
        max_steps = 10,
        print_output=False,
        api_key: Union[str, None] = None,
        base_url: Union[str, None] = None,
    ):
        if base_url:
            self.client = OpenAICompatibleClient(
                model=model,
                tools=tools,
                max_steps=max_steps,
                print_output=print_output,
                base_url=base_url,
                system=system,
            )
        elif model in models["openai_models"]:
            self.client = OpenAIClient(
                model=model,
                system=system,
                tools=tools,
                max_steps=max_steps,
                print_output=print_output,
                api_key=api_key,
            )
        elif model in models["claude_models"]:
            self.client = AnthropicClient(
                model=model,
                system=system,
                tools=tools,
                max_steps=max_steps,
                print_output=print_output,
                api_key=api_key,
            )
        else:
            raise ModelNotSupportedException

    def __call__(
        self,
        prompt: str,
        image: Union[str, None] = None, # base64 string
        current_fn_response = None,
        disable_auto_execution = False,
    ):
        return self.client(
            prompt=prompt,
            image=image,
            current_fn_response=current_fn_response,
            disable_auto_execution=disable_auto_execution,
        )

    @classmethod
    def get_supported_models(cls):
        return models

    def get_messages(self):
        return self.client.messages
    
    def clear_messages(self):
        self.client.messages = [
            message for message in self.client.messages if message["role"]  == "system"
        ]


import os
from typing import Union, List, Dict
from pathlib import Path
from anthropic import Anthropic
from anthropic.types import TextBlock, ToolUseBlock
from toolla.utils import (
    get_image_mime_type,
    load_file_base64,
    build_claude_tool_schema
)
from toolla.exceptions import (
    MessageTooLongException,
    AbortedToolException
)

class AnthropicClient:
    def __init__(
        self,
        model: str,
        system: Union[str, None] = None,
        tools: List[Dict] = [],
        max_steps = 10,
        print_output=False,
        api_key: Union[str, None] = None,
    ):
        self.client = Anthropic(api_key=api_key or os.environ.get("ANTHROPIC_API_KEY"))
        self.model = model
        self.system = system
        self.max_steps = max_steps
        self.messages = []
        self.print_output = print_output

        # TODO change when expanding to other models
        self.max_tokens = 4096
        self.max_chars = 1_000_000

        # Build tool definitions from fns
        if tools:
            # TODO check that docstrings
            # match functions and are valid 
            self.tools = []
            self.tool_fns = {}
            for f in tools:
                self.tool_fns[f.__name__] = f
                tool_dict = build_claude_tool_schema(f)
                self.tools.append(tool_dict)
        else:
            self.tools = []
    
    def __call__(
        self,
        prompt: str,
        image: Union[str, None] = None,
        current_fn_response = None,
        disable_auto_execution = False,
    ):
        message = {
            "role": "user",
            "content": [
                { "type": "text", "text": prompt }
            ]
        }
        self.messages.append(message)
        if image:
            fpath = Path(image)
            mtype = get_image_mime_type(fpath)
            image_string = load_file_base64(fpath) 
            message["content"].append(
                {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": mtype,
                        "data": image_string
                    }
                }
            )
    
        # A bit hacky, should calculate exact length given messages
        while len(str(self.messages)) > self.max_chars:
            self.messages.pop(0)
            if not self.messages:
                raise MessageTooLongException

        # TODO assert if tool choice set, then tools is not None
        response = self.client.messages.create(
            model="claude-3-5-sonnet-20240620",
            max_tokens=self.max_tokens,
            tools=self.tools or [],
            system=self.system or '',
            messages=self.messages,
        )
 
        for content in response.content:
            if isinstance(content, TextBlock):
                self.messages.append({
                    "role": "assistant",
                    "content": content.text,
                })
                if self.print_output:
                    print(f"{content.text}\n")
                if response.stop_reason == 'end_turn':
                    return current_fn_response
            elif isinstance(content, ToolUseBlock):
                fn_inputs = content.input
                # TODO add try catch block and return error
                if disable_auto_execution:
                    print(f"Function {content.name} is about to be called with inputs: {fn_inputs}")
                    user_input = input("Do you want to run this function? (y/n): ")
                    if user_input.lower() not in ['y', 'Y']:
                        print("Function call aborted by user.")
                        raise AbortedToolException
                r = self.tool_fns[content.name](**fn_inputs)
                if len(self.messages) < 2 * self.max_steps:
                    if response.stop_reason == 'tool_use':
                        return self(
                            prompt=f"\nFunction {content.name} was called and returned a value of {r}",
                            current_fn_response=r,
                            disable_auto_execution=disable_auto_execution,
                        )
                else:
                    print("Reached maxiumum number of steps, returning current tool response.")
                    return None
        return None